/*
 * RTEMS configuration and initialization for benchmark application.
 *
 * Written by Michael South, msouth@msouth.org
 * November, 2008
 *
 * This work is in the public domain. I make no warranties of any kind.
 *
 * $ID: $
 */

#include <bsp.h>

#define CONFIGURE_APPLICATION_NEEDS_CONSOLE_DRIVER
#define CONFIGURE_APPLICATION_NEEDS_CLOCK_DRIVER
#define CONFIGURE_RTEMS_INIT_TASKS_TABLE
#define CONFIGURE_LIBIO_MAXIMUM_FILE_DESCRIPTORS 10
#define CONFIGURE_USE_IMFS_AS_BASE_FILESYSTEM

#define CONFIGURE_EXECUTIVE_RAM_SIZE	(512*1024)
#define CONFIGURE_MAXIMUM_SEMAPHORES	20
#define CONFIGURE_MAXIMUM_TASKS		20

#define CONFIGURE_MICROSECONDS_PER_TICK	1000

#define CONFIGURE_INIT_TASK_STACK_SIZE	(10*1024)
#define CONFIGURE_INIT_TASK_PRIORITY	100
#define CONFIGURE_INIT_TASK_INITIAL_MODES (RTEMS_PREEMPT | \
                                           RTEMS_NO_TIMESLICE | \
                                           RTEMS_NO_ASR | \
                                           RTEMS_INTERRUPT_LEVEL(0))

#define CONFIGURE_INIT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rtems.h>
#include <rtems/config.h>
#include <rtems/error.h>

#define BENCH_SAMPLE_SECS       5       /**< Number of seconds per test */

int                     benchpri;       /**< Run benchmarks at this priority */
rtems_interval          tps;            /**< Ticks per second */
int                     phony;          /**< Something keep optimizer happy */
rtems_task_priority     dummy;

/* Iterations / interval stats */
double                  it_ips;         /**< Iterations per second */
double                  it_ipt;         /**< Iterations per tick */
double                  it_usec;        /**< microseconds per iteration */

int                     missed_switch;  /**< # times didn't context switch */


/*****************************************************************************
 * Utilities
 *****************************************************************************/

/** @brief Spin until clock ticks.
 */
static rtems_interval
tickspin() 
{
  rtems_interval  t1, t2;

  for (t1 = rtems_clock_get_ticks_since_boot();
       t1 == (t2 = rtems_clock_get_ticks_since_boot());
       );

  return t2;
}


/** @brief See how many times you can do something in a second / tick / usec.
 *
 * Do @a pf for about 5 seconds, time it, and print the stats.
 */
static void
time_iterations (void (*pf)())
{
  unsigned long  itps;  /* Rough iterations per second */
  unsigned long  itlim; /* How many iterations to do for real */
  unsigned long  it;    /* Current iteration count */
  rtems_interval t1, t2;
  
  /* First, roughly decide how many iterations per second */
  t1 = tickspin();
  t2 = t1 + tps;
  
  for (itps = 0; rtems_clock_get_ticks_since_boot() < t2; ++itps)
    (*pf)();

  /*
   * Now do it for real for a longer period, and without the overhead of
   * clock watching.
   */
  itlim = itps * BENCH_SAMPLE_SECS;
  t1 = tickspin();
  for (it = 0; it < itlim; ++it)
    (*pf)();
  t2 = rtems_clock_get_ticks_since_boot();


  /* Calc stats */
  if (t2 > t1) {
    it_ipt = (double) itlim / (double)(t2 - t1);
    it_ips = it_ipt * tps;
  } else {
    it_ipt = 9999999.9;
    it_ips = 9999999.9;
  }

  if (it_ips > 0.0)
    it_usec = 1000000.0 / it_ips;
  else
    it_usec = 9999999.9;
}


/*****************************************************************************
 * Benchmark -- Empty function call
 *****************************************************************************/

/** @brief Empty function call timing.
 */
static void
bmark_funcall ()
{
  phony = 1;        /* Hopefully keep the optimizer at bay */
}

/** @brief Function to perform empty function call timing.
 */
static void
brun_funcall () 
{
  time_iterations(bmark_funcall);

  printf("%12.2f  %12.2f  %12.3f\n", it_ips, it_ipt, it_usec);
}


/*****************************************************************************
 * Benchmark -- Null context switches
 *****************************************************************************/

/** @brief Perform a context switch.
 */
static void
bmark_nullctxswitch ()
{
  phony = 0;
  (void) rtems_task_wake_after(RTEMS_YIELD_PROCESSOR);
}

/** @brief Function to perform null context switches (yields to self).
 */
static void
brun_nullctxswitch () 
{
  time_iterations(bmark_nullctxswitch);

  printf("%12.2f  %12.2f  %12.3f\n", it_ips, it_ipt, it_usec);
}


/*****************************************************************************
 * Benchmark -- Context switches
 *****************************************************************************/

/** @brief Perform a context switch.
 */
static void
bmark_ctxswitch ()
{
  phony = 0;
  (void) rtems_task_wake_after(RTEMS_YIELD_PROCESSOR);
  if (phony != 1)
      ++missed_switch;
}

/** @brief Aux. task for measuring context switches.
 */
static void
btask_ctxswitchaux () 
{
  for (;;) {
    phony = 1;
    (void) rtems_task_wake_after(RTEMS_YIELD_PROCESSOR);
    if (phony != 0)
        ++missed_switch;
  }
}

/** @brief Function to measure context switches.
 */
static void
brun_ctxswitch () 
{
  rtems_id           aux;
  rtems_status_code  rr;

  /* Create our partner task */
  rr = rtems_task_create(rtems_build_name ('B', 'M', 'k', '2'),
                         benchpri, 4096,
                         RTEMS_PREEMPT | RTEMS_NO_TIMESLICE | RTEMS_NO_ASR,
                         RTEMS_NO_FLOATING_POINT | RTEMS_LOCAL,
                         &aux);
  if (rr != RTEMS_SUCCESSFUL) {
    printf("\nError %s creating auxiliary task.\n", rtems_status_text(rr));
    exit(1);
  }
  rr = rtems_task_start (aux, btask_ctxswitchaux, (rtems_task_argument)&dummy);
  if (rr != RTEMS_SUCCESSFUL) {
    printf("\nError %s starting auxiliary task.\n", rtems_status_text(rr));
    exit(1);
  }

  missed_switch = 0;
  time_iterations(bmark_ctxswitch);
  rtems_task_delete(aux);

  /* Multiply iterations by two to account for the other task's switches. */
  it_ips *= 2.0;
  it_ipt *= 2.0;
  it_usec /= 2.0;
  
  printf("%12.2f  %12.2f  %12.3f\n", it_ips, it_ipt, it_usec);
  if (missed_switch)
    printf("  NOTE: Missed %d context switches.\n", missed_switch);
}


/*****************************************************************************
 * Main
 *****************************************************************************/

rtems_task Init (rtems_task_argument dummy)
{
  char                  linebuf[32];

  printf("\n%s\nBenchmark Test\n\n", _RTEMS_version);

  
  /*
   * Globals
   */
  
  printf("Priority for benchmark tasks [default 120] ?");
  fflush(stdout);
  fgets(linebuf, sizeof(linebuf), stdin);

  benchpri = atoi(linebuf);
  if (benchpri < 2)
    benchpri = 120;

  tps = rtems_clock_get_ticks_per_second();


  /*
   * Execute benchmarks
   */
  rtems_task_set_priority(RTEMS_SELF, benchpri, &dummy);

  printf("Benchmark             Per Second      Per Tick      uSeconds\n");

  printf("%-20s", "Function call");     brun_funcall();
  printf("%-20s", "Yield to self");     brun_nullctxswitch();
  printf("%-20s", "Context switch");    brun_ctxswitch();
  
  exit(0);
}


#include <rtems/confdefs.h>
